import pytest

from pymapd import OperationalError, connect
from pymapd.cursor import Cursor


class TestConnect:

    def test_host_specified(self):
        with pytest.raises(TypeError):
            connect(user='foo')

    def test_raises_right_exception(self):
        with pytest.raises(OperationalError):
            connect(host='localhost', protocol='binary', port=1234)

    def test_close(self, mock_transport, mock_client):
        con = connect(user='user', password='password',
                      host='localhost', dbname='dbname')
        assert con.closed == 0
        con.close()
        assert con.closed == 1

    def test_connect(self, mock_transport, mock_client):
        con = connect(user='user', password='password',
                      host='localhost', dbname='dbname')
        assert mock_client.call_count == 1
        assert con._client.connect.call_args == [
            ('user', 'password', 'dbname')
        ]

    def test_context_manager(self, mock_transport, mock_client):
        con = connect(user='user', password='password',
                      host='localhost', dbname='dbname')
        with con as cur:
            pass

        assert isinstance(cur, Cursor)
        assert con.closed == 0



def no_gpu():
    """Detect if don't have numba and a GPU available"""
    try:
        from numba import cuda

        try:
            cuda.select_device(0)
        except cuda.cudadrv.error.CudaDriverError:
            return True
    except ImportError:
        return True
    return False



from pymapd.exceptions import (_translate_exception, ProgrammingError,
                               DatabaseError)


class TestExceptions:

    def test_invalid_sql_raises(self, invalid_sql):
        result = _translate_exception(invalid_sql)
        assert isinstance(result, ProgrammingError)
        assert 'Validate failed: From line 1,' in result.args[0]

    def test_nonexistant_table_raises(self, nonexistant_table):
        result = _translate_exception(nonexistant_table)
        assert isinstance(result, DatabaseError)
        assert "Exception occurred: Table" in result.args[0]



import os
import pickle
import subprocess
import time

import pytest
import importlib
from thrift.transport import TSocket, TTransport
from thrift.transport.TSocket import TTransportException

from pymapd import connect

ttypes = importlib.import_module("mapd.ttypes")
HERE = os.path.dirname(__file__)

socket = TSocket.TSocket("localhost", 9091)
transport = TTransport.TBufferedTransport(socket)


def _check_open():
    try:
        transport.open()
        return True
    except TTransportException:
        return False


@pytest.fixture(scope='session')
def mapd_server():
    """
    Ensure a mapd server is running.
    """
    if _check_open():
        # already running before pytest started
        pass
    else:
        # not yet running...
        subprocess.check_output(['docker', 'run', '-d',
                                 '-p', '9092:9092', '-p', '9091:9091',
                                 '--name=mapd', 'mapd/mapd:v3.0.0'])
        # yield and stop afterwards?
        assert _check_open()
        # Takes some time to start up. Unfortunately even trying to connect
        # will cause it to hang.
        time.sleep(5)


@pytest.fixture(scope='session')
def con(mapd_server):
    return connect(user="mapd", password='HyperInteractive', host='localhost',
                   port=9091, protocol='binary', dbname='mapd')


def _load_pickle(filepath):
    with open(filepath, 'rb') as f:
        return pickle.load(f)


@pytest.fixture
def colwise():
    return _load_pickle(os.path.join(HERE, "data", "colwise.pkl"))


@pytest.fixture
def rowwise():
    return _load_pickle(os.path.join(HERE, "data", "rowwise.pkl"))


@pytest.fixture
def invalid_sql():
    return _load_pickle(os.path.join(HERE, "data", "invalid_sql.pkl"))


@pytest.fixture
def nonexistant_table():
    return _load_pickle(os.path.join(HERE, "data", "nonexistant_table.pkl"))


@pytest.fixture(scope="session")
def stocks(con):
    drop = 'drop table if exists stocks;'
    c = con.cursor()
    c.execute(drop)
    create = ('create table stocks (date_ text, trans text, symbol text, '
              'qty int, price float, vol float);')
    c.execute(create)
    i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"
    i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"

    c.execute(i1)
    c.execute(i2)


@pytest.fixture
def mock_transport(mocker):
    """A magicmock for pymapd.connection.TTrapnsport.TBufferedTransport"""
    return mocker.patch("pymapd.connection.TTransport.TBufferedTransport")


@pytest.fixture
def mock_client(mocker):
    """A magicmock for pymapd.connection.Client"""
    return mocker.patch("pymapd.connection.Client")


@pytest.fixture
def mock_connection(mock_transport, mock_client):
    """Connection with mocked transport layer, and

    - username='user'
    - password='password'
    - host='localhost'
    - dbname='dbname'
    """
    return connect(user='user', password='password',
                   host='localhost', dbname='dbname')



from collections import namedtuple
from typing import (Iterator, Iterable, List, Optional, Tuple, Any, Union)  # noqa

import six
import mapd.ttypes as T  # noqa

from .exceptions import _translate_exception


Description = namedtuple("Description", ["name", "type_code", "display_size",
                                         "internal_size", "precision", "scale",
                                         "null_ok"])


class Cursor(object):
    """A database cursor."""

    def __init__(self, connection, columnar=True):
        # type: (Any, bool) -> None
        # XXX: supposed to share state between cursors of the same connection
        self.connection = connection
        self.columnar = columnar
        self.rowcount = -1
        self._description = None  # type: Optional[List[Description]]
        self._arraysize = 1
        self._result = None
        self._result_set = None  # type: Optional[Iterator[Any]]

    def __iter__(self):
        # type: () -> Union[List, Iterator]
        if self.result_set is None:
            return iter([])
        return self.result_set

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def description(self):
        # type: () -> Optional[List[Description]]
        """
        Read-only sequence describing columns of the result set.
        Each column is an instance of `Description` describing

        - name
        - type_code
        - display_size
        - internal_size
        - precision
        - scale
        - null_ok

        We only use name, type_code, and null_ok; The rest are always ``None``
        """
        return self._description

    @property
    def result_set(self):
        # type: () -> Optional[Iterator]
        return self._result_set

    @property
    def arraysize(self):
        # type: () -> int
        """The number of rows to fetch at a time with `fetchmany`. Default 1.

        See Also
        --------
        fetchmany
        """
        return self._arraysize

    @arraysize.setter
    def arraysize(self, value):
        # type: (int) -> None
        """Number of items to fetch with :func:`fetchmany`."""
        if not isinstance(value, int):
            raise TypeError("Value must be an integer, got {} instead".format(
                type(value)))
        self._arraysize = value

    def close(self):
        # type: () -> None
        """Close this cursor."""
        # TODO
        pass

    def execute(self, operation, parameters=None):
        # type: (str, tuple) -> Cursor
        """Execute a SQL statement.

        Parameters
        ----------
        operation : str
            A SQL query
        parameters : tuple
            Parameters to substitute into ``operation``.

        Returns
        -------
        self : Cursor
        """
        if parameters is not None:
            raise NotImplementedError
        self.rowcount = -1
        try:
            result = self.connection._client.sql_execute(
                self.connection._session, operation,
                column_format=self.columnar,
                nonce=None, first_n=-1)
        except T.TMapDException as e:
            six.raise_from(_translate_exception(e), e)
        self._description = _extract_description(result.row_set.row_desc)
        if self.columnar:
            try:
                self.rowcount = len(result.row_set.columns[0].nulls)
            except IndexError:
                pass
        else:
            self.rowcount = len(result.row_set.rows)
        self._result_set = make_row_results_set(result)
        self._result = result
        return self

    def executemany(self, operation, parameters=None):
        # type: (str, Iterable) -> None
        pass

    def fetchone(self):
        # type: () -> Optional[Any]
        """Fetch a single row from the results set"""
        try:
            return next(self.result_set)
        except StopIteration:
            return None

    def fetchmany(self, size=None):
        # type: (Optional[int]) -> Iterable[Any]
        """Fetch ``size`` rows from the results set."""
        if size is None:
            size = self.arraysize
        results = [self.fetchone() for _ in range(size)]
        return [x for x in results if x is not None]

    def fetchall(self):
        # type: () -> Any
        return list(self)

    def setinputsizes(self, sizes):
        # type: (int) -> None
        pass

    def setoutputsizes(self, size, column=None):
        # type: (int, Optional[Any]) -> None
        pass


# -----------------------------------------------------------------------------
# Result Sets
# -----------------------------------------------------------------------------

def make_row_results_set(data):
    # type: (T.QueryResultSet) -> Iterator[Tuple]
    """
    Build a results set of python objects.

    Parameters
    ----------
    data : QueryResultSet

    Returns
    -------
    results : Iterator[tuple]
    """
    if is_columnar(data):
        nrows = len(data.row_set.columns[0].nulls)
        ncols = len(data.row_set.row_desc)
        columns = [_extract_col_vals(desc, col)
                   for desc, col in zip(data.row_set.row_desc,
                                        data.row_set.columns)]
        for i in range(nrows):
            yield tuple(columns[j][i] for j in range(ncols))
    else:
        for row in data.row_set.rows:
            yield tuple(_extract_row_val(desc, val)
                        for desc, val in zip(data.row_set.row_desc, row.cols))


def is_columnar(data):
    # type: (T.TQueryResult) -> bool
    return data.row_set.is_columnar


_typeattr = {
    'SMALLINT': 'int',
    'INT': 'int',
    'BIGINT': 'int',
    'TIME': 'int',
    'TIMESTAMP': 'int',
    'DATE': 'int',
    'BOOL': 'int',
    'FLOAT': 'real',
    'DECIMAL': 'real',
    'DOUBLE': 'real',
    'STR': 'str',
}


def _extract_row_val(desc, val):
    # type: (T.TColumnType, T.TDatum) -> Any
    typename = T.TDatumType._VALUES_TO_NAMES[desc.col_type.type]
    return getattr(val.val, _typeattr[typename] + '_val')


def _extract_col_vals(desc, val):
    # type: (T.TColumnType, T.TColumn) -> Any
    typename = T.TDatumType._VALUES_TO_NAMES[desc.col_type.type]
    return getattr(val.data, _typeattr[typename] + '_col')


def _extract_description(row_desc):
    # type: (List[T.TColumnType]) -> List[Description]
    """
    Return a tuple of (name, type_code, display_size, internal_size,
                       precision, scale, null_ok)

    https://www.python.org/dev/peps/pep-0249/#description
    """
    return [Description(col.col_name, col.col_type.type,
                        None, None, None, None,
                        col.col_type.nullable)
            for col in row_desc]



"""
https://www.python.org/dev/peps/pep-0249/#type-objects
"""
import datetime
import six
import time
from typing import List, Any  # noqa

from mapd import MapD

if six.PY2:
    memoryview = buffer  # noqa


T = MapD.TDatumType


class DataType(object):

    def __init__(self, matches):
        # type: (List[Any]) -> None
        self._matches = set(matches)

    def __eq__(self, other):
        # type: (Any) -> bool
        return other in self._matches

    def __ne__(self, other):
        # type: (Any) -> bool
        return not (self == other)

    def __hash__(self):
        # type: () -> int
        return hash(tuple(self._matches))


Binary = memoryview
Date = datetime.date
Time = datetime.time
Timestamp = datetime.datetime

BINARY = DataType([])
STRING = DataType([T.STR])
NUMBER = DataType([T.SMALLINT, T.INT, T.BIGINT, T.FLOAT, T.DECIMAL, T.DOUBLE,
                   T.BOOL])
DATETIME = DataType([T.DATE, T.TIME, T.TIMESTAMP])
ROWID = DataType([])


def DateFromTicks(ticks):
    # type: (int) -> Date
    return Date(*time.localtime(ticks)[:3])


def TimeFromTicks(ticks):
    # type: (int) -> Time
    return Time(*time.localtime(ticks)[3:6])


def TimestampFromTicks(ticks):
    # type: (int) -> Timestamp
    return Timestamp(*time.localtime(ticks)[:6])



"""
This script generates the thrift structures used for unit testing pymapd.
It requires a mapd server to be running on localhost:9091 with the default
username, password, and database name.

.. code-block:: console

    python scripts/generate_thrift_structs.py

We run through a series of sql statements using the bindings generated by
thrift. The return values are serialized to python files, which are then
read in and executed by the test suite.
"""
import os
import pickle
from thrift.protocol import TBinaryProtocol
from thrift.transport import TSocket
from thrift.transport import TTransport
from mapd import MapD
from mapd.ttypes import TMapDException

HERE = os.path.dirname(__file__)
DEST = os.path.join(os.path.dirname(HERE), "tests", "data")


def get_client(host_or_uri, port):
    socket = TSocket.TSocket(host_or_uri, port)
    transport = TTransport.TBufferedTransport(socket)
    protocol = TBinaryProtocol.TBinaryProtocol(transport)

    client = MapD.Client(protocol)
    transport.open()
    return client


def write(obj, name):
    with open(os.path.join(DEST, name), 'wb') as f:
        pickle.dump(obj, f, protocol=2)


def main():

    db_name = 'mapd'
    user_name = 'mapd'
    passwd = 'HyperInteractive'
    hostname = 'localhost'
    portno = 9091

    client = get_client(hostname, portno)
    session = client.connect(user_name, passwd, db_name)

    drop = 'drop table if exists stocks;'
    client.sql_execute(session, drop, True, None, -1)
    create = ('create table stocks (date_ text, trans text, symbol text, '
              'qty int, price float, vol float);')
    client.sql_execute(session, create, True, None, -1)

    i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"
    i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"
    client.sql_execute(session, i1, True, None, -1)
    client.sql_execute(session, i2, True, None, -1)
    select = "select * from stocks;"
    colwise = client.sql_execute(session, select, True, None, -1)
    rowwise = client.sql_execute(session, select, False, None, -1)

    write(rowwise, "rowwise.pkl")
    write(colwise, "colwise.pkl")
    # Invalid SQL
    try:
        client.sql_execute(session, "select it;", True, None, -1)
    except TMapDException as e:
        write(e, "invalid_sql.pkl")

    # Valid SQL, non-existant table
    try:
        client.sql_execute(session, "select fake from not_a_table;", True,
                           None, -1)
    except TMapDException as e:
        write(e, "nonexistant_table.pkl")

    # valid table, non-existant column
    try:
        client.sql_execute(session, "select fake from stocks;", True,
                           None, -1)
    except TMapDException as e:
        write(e, "nonexistant_column.pkl")

    client.disconnect(session)


if __name__ == '__main__':
    main()



"""
Tests that rely on a server running
"""
import pytest
from mapd.ttypes import TMapDException, TGpuDataFrame

from pymapd import connect, ProgrammingError, DatabaseError
from pymapd.cursor import Cursor

from .utils import no_gpu

# XXX: Make it hashable to silence warnings; see if this can be done upstream
# This isn't a huge deal, but our testing context mangers for asserting
# exceptions need hashability
TMapDException.__hash__ = lambda x: id(x)


@pytest.mark.usefixtures("mapd_server")
class TestIntegration:

    @pytest.mark.parametrize('protocol', [
        pytest.mark.skip(reason="Hangs waiting to hear back")('http'),
        'binary'])
    def test_conenct(self, protocol):
        con = connect(user="mapd", password='HyperInteractive',
                      host='localhost', port=9091, protocol=protocol,
                      dbname='mapd')
        assert con is not None
        assert protocol in repr(con)

    def test_invalid_sql(self, con):
        with pytest.raises(ProgrammingError) as r:
            con.cursor().execute("select it;")
        r.match("Column 'it' not found in any table")

    def test_nonexistant_table(self, con):
        with pytest.raises(DatabaseError) as r:
            con.cursor().execute("select it from fake_table;")
        r.match("Table 'FAKE_TABLE' does not exist")

    def test_connection_execute(self, con):
        result = con.execute("drop table if exists FOO;")
        result = con.execute("create table FOO (a int);")
        assert isinstance(result, Cursor)

    @pytest.mark.skipif(no_gpu, reason="No GPU available")
    def test_select_ipc(self, con, stocks):
        result = con.select_ipc("select qty, price from stocks")
        assert isinstance(result, TGpuDataFrame)

    @pytest.mark.skipif(no_gpu, reason="No GPU available")
    def test_select_ipc_gpu(self, con, stocks):
        result = con.select_ipc_gpu("select qty, price from stocks")
        assert isinstance(result, TGpuDataFrame)
